﻿//#define USE_DUNGEONBUDDY_DLL

using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;
using System.Windows.Media;
using CommonBehaviors.Actions;
using MainDev.RemoteASM.Handlers;
using Styx;
using Styx.Common.Helpers;
using Styx.CommonBot;
using Styx.CommonBot.Frames;
using Styx.Helpers;
using Styx.Patchables;
using Styx.Pathing;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Styx.WoWInternals.World;
using Action = Styx.TreeSharp.Action;

#if USE_DUNGEONBUDDY_DLL
using Bots.DungeonBuddyDll;
using Bots.DungeonBuddyDll.Profiles;
using Bots.DungeonBuddyDll.Attributes;
using Bots.DungeonBuddyDll.Helpers;

namespace Bots.DungeonBuddyDll.Dungeon_Scripts.Mists_of_Pandaria
#else
    using Bots.DungeonBuddy.Profiles;
    using Bots.DungeonBuddy.Attributes;
    using Bots.DungeonBuddy.Helpers;
namespace Bots.DungeonBuddy.Dungeon_Scripts.Mists_of_Pandaria
#endif
{
    public class TerraceOfEndlessSpring : Dungeon
    {
        #region Overrides of Dungeon

        public override uint DungeonId
        {
            get { return 526; }

        }

        public override WoWPoint Entrance
        {
            get { return new WoWPoint(958.4459, -51.12228, 513.1506); }
        }

        public override void RemoveTargetsFilter(List<WoWObject> units)
        {
            units.RemoveAll(
                ret =>
                {
                    var unit = ret.ToUnit();
                    if (unit != null)
                    {
                        if (unit.Entry == LeiShiId && unit.HasAura("Protect"))
                            return true;
                    }
                    return false;
                });
        }

        public override void IncludeTargetsFilter(List<WoWObject> incomingunits, HashSet<WoWObject> outgoingunits)
        {
            foreach (var obj in incomingunits)
            {
                var unit = obj as WoWUnit;
                if (unit != null)
                {
                    if (unit.Entry == TerrorSpawnId && unit.Distance < 80)
                        outgoingunits.Add(unit);
                }
            }
        }

        public override void WeighTargetsFilter(List<Targeting.TargetPriority> units)
        {
            foreach (var priority in units)
            {
                var unit = priority.Object as WoWUnit;
                if (unit != null)
                {
                    if (unit.Entry == TerrorSpawnId)
                        priority.Score += 5000;
                }
            }
        }

        public override void OnEnter()
        {
            if (Me.IsTank())
            {
                Alert.Show(
                    "Tanking Not Supported",
                    string.Format(
                        "Tanking is not supported in the {0} script. If you wish to stay in raid and play manually then press 'Continue'. Otherwise you will automatically leave raid.",
                        Name),
                    30,
                    true,
                    true,
                    null,
                    () => Lua.DoString("LeaveParty()"),
                    "Continue",
                    "Leave");
            }
            else
            {
                Alert.Show(
                    "Do Not AFK",
                    "It is highly recommended you do not afk while in a raid and be prepared to intervene if needed in the event something goes wrong or you're asked to perform a certain task.",
                    20,
                    true,
                    false,
                    null,
                    null,
                    "Ok");
            }
        }

        #endregion

        private static LocalPlayer Me
        {
            get { return StyxWoW.Me; }
        }

        [EncounterHandler(0)]
        public Composite RootEncounter()
        {
            return new PrioritySelector(
                );
        }

        #region Protectors of the Endless

        WoWPoint _protectersCenterLoc = new WoWPoint(-1017.489, -3048.979, 12.82353);

        // Lightning storm. - just heal through it. 
        [EncounterHandler(60583, "Protector Kaolan")]
        public Composite ProtectorKaolanEncounter()
        {
            WoWUnit boss = null;
            return new PrioritySelector(ctx => boss = ctx as WoWUnit);
        }

        [EncounterHandler(60585, "Elder Regail")]
        public Composite ElderRegailEncounter()
        {
            WoWUnit boss = null;
            return new PrioritySelector(ctx => boss = ctx as WoWUnit);
        }

        [EncounterHandler(60586, "Elder Asani")]
        public Composite ElderAsaniEncounter()
        {
            WoWUnit boss = null;
            return new PrioritySelector(ctx => boss = ctx as WoWUnit);
        }

        #endregion


        #region Tsulong

        [EncounterHandler(62442, "Tsulong")]
        public Composite TsulongEncounter()
        {
            WoWUnit boss = null;
            return new PrioritySelector(ctx => boss = ctx as WoWUnit);
        }

        #endregion

        #region Lei Shi

        private const uint LeiShiId = 62983;
        const uint HiddenLeiShiId = 63099;
        const int GetAwayId = 123461;

        [EncounterHandler(62983, "Lei Shi")]
        public Composite LeiShiEncounter()
        {
            WoWUnit boss = null;

            return new PrioritySelector(ctx => boss = ctx as WoWUnit,
                                                    new Decorator(ctx => boss.CastingSpellId == GetAwayId && boss.Distance > 30,
                                                        new Action(ctx => Navigator.PlayerMover.MoveTowards(boss.Location)))
                                            );
        }

        [EncounterHandler(62983, "Lei Shi", Mode = CallBehaviorMode.CurrentBoss)]
        public Composite HiddenLeiShiEncounter()
        {
            WoWUnit boss = null;
            WoWUnit hiddenLeiShi = null;

            return new PrioritySelector(ctx =>
            {
                hiddenLeiShi = ObjectManager.GetObjectsOfType<WoWUnit>().FirstOrDefault(u => u.Entry == HiddenLeiShiId);
                return boss = ctx as WoWUnit;
            },
                                            new Decorator(ctx => hiddenLeiShi != null,
                                                new PrioritySelector(
                                                    new Decorator(ctx => hiddenLeiShi.Distance > 4, new Action(ctx => Navigator.MoveTo(hiddenLeiShi.Location))),
                                                    new Decorator(ctx => hiddenLeiShi.Distance <= 4,
                                                        new PrioritySelector(
                                                            new Decorator(ctx => Me.IsCasting,new ActionAlwaysSucceed()),
                                                            Cast("Divine Storm"),
                                                            Cast("Thunder Clap"),
                                                            Cast("Arcane Explostion"),
                                                            Cast("Rain of Fire")
                                                            ))
                                                    ))
                                            );
        }

        Composite Cast(string spell)
        {
            return new Action(
                ctx =>
                {
                    if (SpellManager.Cast(spell))
                        return RunStatus.Success;
                    return RunStatus.Failure;
                });
        }

        #endregion

        #region Sha of Fear

        const uint TerrorSpawnId = 61034;
        [EncounterHandler(60999, "Sha of Fear")]
        public Composite ShaOfFearEncounter()
        {
            WoWUnit boss = null;
            WoWUnit myTarget = null;

            return new PrioritySelector(ctx => boss = ctx as WoWUnit,
                    ScriptHelpers.GetBehindUnit(ctx => (myTarget = Me.CurrentTarget) != null && myTarget.Entry == TerrorSpawnId && myTarget.IsFacing(Me), () => myTarget)
                );
        }

        #endregion


    }
}