﻿//#define USE_DUNGEONBUDDY_DLL

using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;
using System.Windows.Media;
using CommonBehaviors.Actions;
using MainDev.RemoteASM.Handlers;
using Styx;
using Styx.Common.Helpers;
using Styx.CommonBot;
using Styx.CommonBot.Frames;
using Styx.Helpers;
using Styx.Patchables;
using Styx.Pathing;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Styx.WoWInternals.World;
using Action = Styx.TreeSharp.Action;
#if USE_DUNGEONBUDDY_DLL
using Bots.DungeonBuddyDll;
using Bots.DungeonBuddyDll.Profiles;
using Bots.DungeonBuddyDll.Attributes;
using Bots.DungeonBuddyDll.Helpers;

namespace Bots.DungeonBuddyDll.Dungeon_Scripts.Mists_of_Pandaria
#else
    using Bots.DungeonBuddy.Profiles;
    using Bots.DungeonBuddy.Attributes;
    using Bots.DungeonBuddy.Helpers;
namespace Bots.DungeonBuddy.Dungeon_Scripts.Mists_of_Pandaria
#endif
{
    public class TheDreadApproach : Dungeon
    {
        #region Overrides of Dungeon

        public override uint DungeonId
        {
            get { return 529; }
        }

        public override WoWPoint Entrance
        {
            get { return new WoWPoint(165.4734, 4069.003, 255.8758); }
        }

        public override void RemoveTargetsFilter(List<WoWObject> units)
        {
            units.RemoveAll(
                ret =>
                {
                    var unit = ret.ToUnit();
                    if (unit != null)
                    {
                        if (unit.Entry == GaralonsLegId && Me.IsMelee())
                        {
                            if (!Navigator.CanNavigateFully(Me.Location, unit.Location))
                                return true;
                            // have melee ignore the legs that are by the wall bcause of potential issues getting to them e.g. having to run in front of boss or stand in the puddles.
                            var transport = unit.Transport;
                            if (transport != null)
                            {
                                var transportLoc = transport.Location;
                                var travelingClockwise = !_garalonRoomCenter.IsPointLeftOfLine(
                                    transportLoc, transportLoc.RayCast(WoWMathHelper.NormalizeRadian(transport.Rotation), 10));
                                var relativeLoc = unit.RelativeLocation;
                                return relativeLoc.Y < 0 && !travelingClockwise || relativeLoc.Y > 0 && travelingClockwise;
                            }
                        }
                    }
                    return false;
                });
        }

        public override void IncludeTargetsFilter(List<WoWObject> incomingunits, HashSet<WoWObject> outgoingunits)
        {
            foreach (var obj in incomingunits)
            {
                var unit = obj as WoWUnit;
                if (unit != null) { }
            }
        }

        public override void WeighTargetsFilter(List<Targeting.TargetPriority> units)
        {
            foreach (var priority in units)
            {
                var unit = priority.Object as WoWUnit;
                if (unit != null)
                {
                    if (unit is WoWPlayer)
                        priority.Score += 1000; // dps MCed players.
                    else if (unit.Entry == GaralonsLegId)
                    {
                        priority.Score = 5000 - unit.Distance;
                    }
                }
            }
        }

        public override void OnEnter()
        {
            if (Me.IsTank())
            {
                Alert.Show(
                    "Tanking Not Supported",
                    string.Format(
                        "Tanking is not supported in the {0} script. If you wish to stay in raid and play manually then press 'Continue'. Otherwise you will automatically leave raid.",
                        Name),
                    30,
                    true,
                    true,
                    null,
                    () => Lua.DoString("LeaveParty()"),
                    "Continue",
                    "Leave");
            }
            else
            {
                Alert.Show(
                    "Do Not AFK",
                    "It is highly recommended you do not afk while in a raid and be prepared to intervene if needed in the event something goes wrong or you're asked to perform a certain task.",
                    20,
                    true,
                    false,
                    null,
                    null,
                    "Ok");
            }
        }

        #endregion

        private const uint ZephyrId = 63599;
        private const uint TrashVersionTempestStalkerId = 64373;

        private static LocalPlayer Me
        {
            get { return StyxWoW.Me; }
        }


        [EncounterHandler(0)]
        public Composite RootEncounter()
        {
            AddAvoidObject(
                ctx => !Me.IsCasting,
                7,
                u => u.Entry == ZephyrId,
                o =>
                {
                    var loc = o.Location;
                    return Me.Location.GetNearestPointOnSegment(loc, loc.RayCast(WoWMathHelper.NormalizeRadian(o.Rotation), 15));
                });
            return new PrioritySelector();
        }

        #region Imperial Vizier Zor'lok

        // get in noice canceling shield.
        // ingore sonic ring, too many
        // bunch up for aoe heals.
        // don't get traped on a platform when boss moves to center.

        private const int NoiseCancellingAreaTriggerSpellId = 122731;

        [EncounterHandler(62980, "Imperial Vizier Zor'lok")]
        public Composite ImperialVizierZorlokEncounter()
        {
            WoWUnit boss = null;
            WoWAreaTrigger noiceCancellingArea = null;
            WoWPoint raidLoc = WoWPoint.Zero;

            return new PrioritySelector(
                ctx =>
                {
                    noiceCancellingArea =
                        ObjectManager.GetObjectsOfType<WoWAreaTrigger>().Where(a => a.SpellId == NoiseCancellingAreaTriggerSpellId).OrderBy(a => a.DistanceSqr).
                            FirstOrDefault();
                    raidLoc = ScriptHelpers.GetGroupCenterLocation();
                    return boss = ctx as WoWUnit;
                },
                // get the Noice Cancelling buff.
                new Decorator(ctx => noiceCancellingArea != null && noiceCancellingArea.Distance > 3.5, new Action(ctx => Navigator.MoveTo(noiceCancellingArea.Location))),
                // stack up on raid location.
                new Decorator(
                    ctx => raidLoc.Distance(Me.Location) > 10 && !Me.HasAura("Noise Cancelling") && Me.IsRange(),
                    new Action(ctx => Navigator.MoveTo(noiceCancellingArea.Location))));
        }

        #endregion

        #region Blade Lord Ta'yak

        private const uint TempestStalkerId = 62908;
        private const uint UnleashedWest1TornadoId = 63278;
        private const uint UnleashedWest2TornadoId = 63299;
        private const uint UnleashedWest3TornadoId = 63300;

        private const uint UnleashedEast2TornadoId = 63302;
        private const uint UnleashedEast3TornadoId = 63303;
        private const uint GaleWindId = 63292;
        private readonly uint[] _unleashedTornadoIds = new[] { UnleashedWest1TornadoId, UnleashedWest2TornadoId, UnleashedWest3TornadoId, UnleashedEast2TornadoId, UnleashedEast3TornadoId };

        // dodge the whirlwinds id 62908 (stationary) and 63299 travels in straight line
        // group up on Unseen strike target.(red mark)
        [EncounterHandler(62543, "Blade Lord Ta'yak")]
        public Composite BladeLordTayakEncounter()
        {
            WoWUnit boss = null;
            WoWPlayer unseenStrikeTarget = null;

            AddAvoidObject(ctx => true, 5, TempestStalkerId);
            AddAvoidObject(ctx => true, 4, GaleWindId);
            AddAvoidObject(ctx => true, 4, _unleashedTornadoIds);
            AddAvoidObject(
                ctx => true,
                4,
                o => _unleashedTornadoIds.Contains(o.Entry) && o.Distance < 15,
                o =>
                {
                    var loc = o.Location;
                    return Me.Location.GetNearestPointOnLine(loc, loc.RayCast(WoWMathHelper.NormalizeRadian(o.Rotation), 20));
                });

            return new PrioritySelector(
                ctx =>
                {
                    unseenStrikeTarget = Me.RaidMembers.FirstOrDefault(r => r.HasAura("Unseen Strike"));
                    return boss = ctx as WoWUnit;
                },
                new Decorator(ctx => unseenStrikeTarget != null && unseenStrikeTarget.Distance > 4, new Action(ctx => Navigator.MoveTo(unseenStrikeTarget.Location))));
        }

        #endregion

        #region Garalon

        private const uint GaralonsLegId = 63053;
        private const uint GaralonId = 63191;
        private const uint PheromoneTrailId = 63021;
        private readonly WoWPoint _garalonRoomCenter = new WoWPoint(-1926.08, 475.6746, 470.9576);
        // dps the legs when alive
        // don't go under boss,
        // stay out of the puddles.
        // don't stand in front.

        [EncounterHandler(63191, "Garalon", BossRange = 130)]
        public Composite GaralonEncounter()
        {
            WoWUnit garalon = null;
            var bottomOfStepLoc = new WoWPoint(-1968.789, 476.0225, 470.9576);
            // dont stand under boss.
            AddAvoidObject(ctx => true, () => _garalonRoomCenter, 45, 19, GaralonId);
            // puddles left by Pheromones. only avoid those within 30 yards because the sheer number of these cause a performance hit 
            AddAvoidObject(ctx => true, 4, u => u.Entry == PheromoneTrailId && u.DistanceSqr <= 900);
            // avoid players with Pheromones.
            AddAvoidObject(ctx => true, 10, o => o is WoWPlayer && o.ToPlayer().HasAura("Pheromones") && o.Guid != Me.Guid);
            return new PrioritySelector(
                ctx => garalon = ctx as WoWUnit,
                // avoid  standing in front.
                ScriptHelpers.CreateAvoidUnitAnglesBehavior(ctx => garalon.Combat && !Me.IsCasting && garalon.Distance < 50, () => garalon, new ScriptHelpers.AngleSpan(0, 130)),
                // get off the stairs.
                new Decorator(ctx => !garalon.Combat && Me.Z > 472 && ScriptHelpers.GetGroupCenterLocation().Z < 472,
                    new Action(ctx => Navigator.MoveTo(ScriptHelpers.GetGroupCenterLocation())))
                );
        }

        #endregion
    }
}